function [chl,chl_g,verr] = SGD(bname,fname, drop, lr, mom, chl_in, chl_gin)

true = load('-mat','data/real_illum.mat');
true = true.real_rgb;
true = bsxfun(@rdivide,true,sqrt(sum(true.^2,2)));

tdata = load('-mat',bname);

lprior = tdata.lprior * tdata.beta;

tidx = tdata.train_idx;
TSHFT = 6; VSHFT = 7;

if length(chl_in) == 0
  chl = tdata.chl * tdata.alpha;
  chl_g = zeros(size(chl));
  do_test(tdata,chl,fname,true,tidx,VSHFT);
else
  chl = chl_in;
  chl_g = chl_gin;
end;

lut = uint32(tdata.lut); nillum = size(lut,2);
Q = sqrt(size(chl,1)); LQ = size(chl,2);
ul = tdata.ul; chl_sz = prod(size(chl));

dispint = ceil(length(tidx)/2);

for shft = 0:TSHFT
  
  tidx2 = tidx(mod([1:length(tidx)]+shft-1,length(tidx))+1);
  Alut = mkLUT(Q,true(tidx,:) ./ true(tidx2,:));
  
  for i = 1:length(tidx)
    id = tidx(i); id2 = tidx2(i);
    llut = Alut(:,i);
    
    p = load('-mat',sprintf(fname,id));
    q = uint32(p.q); V0 = p.v;
        
    % Dropout
    didx = dropout(drop(1),drop(2),p.mask);
    q = (q(didx)-1)*Q*Q; V0 = uint32(llut(V0(didx)));
    
    % Forward
    cost = forward(V0,q,lut,chl) + lprior; 
    cost = cost - max(cost); 
    cost = exp(cost); cost = cost / sum(cost);
    
    
    % Compute errs
    errs = acosd(min(1,ul*true(id2,:)'));
    L = sum(cost.*errs);
    
    fprintf('\r S: %d,%03d Loss=%.4f   ',shft,i,L);

    if mod(i,dispint) == 1
      fprintf('\n');
    end;
    
    % Backprop
    gL = -cost.*(errs - L) / length(q);
    chl_gi = reshape(backward(V0,q,lut,gL,chl_sz),size(chl));
  
    chl_g = chl_g*mom + chl_gi;
    chl = chl + chl_g*lr;
  end;
  if mod(length(tidx),dispint) ~= 1
    fprintf('\n');
  end;
end;

verr = do_test(tdata,chl,fname,true,tidx,VSHFT);

% Select subset of patches from image
function sidx = dropout(drop,psz,msk)
lngth = prod(size(msk));
sidx = randperm(lngth,round(drop*lngth/psz/psz));
l2 = zeros(size(msk));
l2(sidx) = 1;
sidx = find(conv2(ones(psz,1),ones(psz,1),l2,'same') >= 1 & msk == 1);


% Compute error on validation set
function verr = do_test(tdata,chl,fname,true,vidx,shft)

lut = uint32(tdata.lut); nillum = size(lut,2);
Q = sqrt(size(chl,1)); LQ = size(chl,2);

err = 0;

vidx2 = vidx(mod([1:length(vidx)]+shft-1,length(vidx))+1);
Alut = mkLUT(Q,true(vidx,:) ./ true(vidx2,:));
for i = 1:length(vidx)

  id = vidx(i); id2 = vidx2(i);
  
  p = load('-mat',sprintf(fname,id));
  q = uint32(p.q(p.mask == 1)-1)*Q*Q; 
  V0 = p.v(p.mask == 1);
    
  llut = Alut(:,i); V0 = uint32(llut(V0));

  cost = forward(V0,q,lut,chl);
  %cost = zeros(nillum,1);
  %for j = 1:nillum
  %  V = lut(:,j); V = V(V0); V = V + q;
  %  cost(j) = sum(chl(V))/length(q);
  %end;
  
  cost = cost + tdata.lprior*tdata.beta;
  cost = cost - max(cost);
  cost = exp(cost); cost = cost / sum(cost);

  est = sum(bsxfun(@times,tdata.ul,cost),1);
  est = est / sqrt(sum(est.^2));
  
  err = err + acosd(min(1,sum(est.*true(id2,:),2)));
  fprintf('\r Val Error = %.4f [%03d]',err / i, i);
end;
fprintf('\n');
verr = err / length(vidx);
